function [optimumNumCentroids, minErr] = optimizeNumberOfCentroids( ...
  maxNumCentroids, simMat, classLabels)
%optimizeNumberOfCentroids - cross-validate number of centroids.
% Cross validate the number of centroids to be used in a
% multi-nearest- centroid classifier for a two-class problem. Uses the
% cross-validation strategy in HTF, pg 216.
%
% USAGE
% [optimumNumCentroids, minErr] = optimizeNumberOfCentroids( ...
%     maxNumCentroids, simMat, classLabels);
% maxNumCentroids: maximum number of centroidsper class to test. The
%   function will cross-validate 1:maxNumCentroids.
% simMat: square matrix of pairwise similarities.
% classLabels: vector of numeric class labels correpsonding to each row in
%   simMat. Only two classes are supported by this function, but their
%   labels can be any numeric value.
% optimumNumCentroids: 2-by-1 vector containing the optimum (minimum error)
%   number of centroids for each class
% minErr: classification error corresponding to the optimum number of
%   centroids.
%
% See also computeMultiCentroidModels(), classifyMultiNearestCentroid().

% Written by Luca Cazzanti, engineer.
% Applied Physics Laboratory, University of Washington, Feb 2007.
%$Id:$
%$DateTime:$

[nData, notUsed] = size(simMat); %#ok

errMat = 10e10 * ones(maxNumCentroids);
nCentroidsVec = 1:maxNumCentroids;
for iCentroid = nCentroidsVec(:).'
  for jCentroid = nCentroidsVec(:).'
    
  classEst = 10e10 * ones(nData,1);
  nCentroidsPerClass = [iCentroid jCentroid];
  for iData = 1:nData
    testPointSim = simMat( iData, :);
    trainSimMat = simMat;
    trainSimMat(iData, :) = [];
    trainSimMat(:, iData) = [];
    trainClassLabels = classLabels;
    trainClassLabels(iData) = [];
    
    initCentroidIdx = initMultiCentroids(nCentroidsPerClass, trainClassLabels);
    modelParams = computeMultiCentroidModels(trainSimMat, ...
      trainClassLabels, initCentroidIdx, 'maxSumSim'); 
   
    classEst(iData) = classifyMultiNearestCentroid(testPointSim, modelParams, ...
     'similarity');
  end %for iData
  
  errMat(iCentroid, jCentroid) = sum( classEst(:) ~= classLabels(:) );
  end %for jCentroid
end %for iCentroid

[minErr1, bestNumIdx1] = min(errMat);
[minErr, bestNumIdx2] = min(minErr1);
bestNumCentroids1 = nCentroidsVec(bestNumIdx1(bestNumIdx2));
bestNumCentroids2 = nCentroidsVec(bestNumIdx2); 
optimumNumCentroids = [bestNumCentroids1 bestNumCentroids2];

return