function runMultiCentroid(flipProbs, fileName, simFunction)
%runMultiCentroid - multi-centroid simulation.
% Run the perturbed centroids simulation for multi-centroid classes
% This is a driver function that calls the appropriate scripts, and saves
% the results to a mat file. Some parameters are provided as inputs, others
% must be set within this function. 
%
% USAGE 
% runMultiCentroid(flipProbs, fileName, simFunction);
% flipProbs: vector of probability that a feature is flipped; its length is
%     equal to the number of classes.
% fileName: name of file where results are to be saved.
% simFunction: either 'Hamming', 'vdm', or 'resEnt';
% 
% EXAMPLE
% flipRobs = [1/3 1/30]; % two classes
% fileMane = 'multi_centroid_results.mat';
% simFunction = 'Hamming';
% runMultiCentroid(flipProbs, fileName, simFunction);

% Written by Luca Cazzanti, engineer
% Applied Physics Laboratory, University of Washington, Dec 2006.
%$Id:$
%$DateTime:$
%$Author:$


%% Which classifiers are we testing?
doSda = true;
doLocalSda = true;
doMixtureSda = true;
doNNSimSda = true;
doNearestCentroid = true;
doMultiNearestCentroid = true;
doLocalNearestCentroid = true;
doNearestNeighbor = true;
doNaiveBayes = true;


%% Set simulation parameters

% Data parameters
%nFeaturesVec = [2 4 8 12 25 40 50 75 100 125 150 175 200];
nFeaturesVec = [2 100];
nTrain = 100;
nTest = 1000;
nCentroidsPerClass = 1;
nIterations = 2;

% Neighborhood sizes to try 
maxNeighbors = 100;
%defaultNeighborhoodSizes = [1:20 30:10:maxNeighbors];
defaultNeighborhoodSizes = [1 9];

% Max number of centroids to try (only for multi-centroid nearest centroid
% classifier)
nCentroidsToTry = 3;

% Pre-allocate matrices here: if not all classifiers are tested, some of
% these will remain unused; that's fine.
totalResultSda = [];
totalResultLocalSda = [];
totalResultLocalNC = [];
totalResultNC = [];
totalResultMultiNC = [];
totalResultPsvm = [];
totalResultRawPsvm = [];
totalResultNN = [];
totalResultNaiveBayes = [];
totalResultMixtureSda = [];
totalResultNNSimSda = [];

testResultLocalSda = [];
testResultLocalNC = [];
testResultNN = [];
testResultSda = [];
testResultNC = [];
testResultMultiNC = [];
testResultNaiveBayes = [];
testResultMixtureSda = [];
testResultNNSimSda = [];

nFeaturesToTry = length(nFeaturesVec);
optimumSizesLocalSda = zeros(nIterations, nFeaturesToTry);
optimumSizesLocalNC = zeros(nIterations, nFeaturesToTry);
optimumSizesNN = zeros(nIterations, nFeaturesToTry);
optimumNumberOfCentroids = cell(nIterations, nFeaturesToTry);
optimumNumberOfCentroidsMixtureSda = cell(nIterations, nFeaturesToTry);


%% Run simulations 

% Repeat test many times to get more robust statistical estimates
for iIter = 1:nIterations
  
  % Initialize to large number for easy bug tracking
  minTrainErrorLocalSda = 1e10 * ones(nFeaturesToTry,1);
  minTrainErrorLocalNC = 1e10 * ones(nFeaturesToTry,1);
  minTrainErrorNN = 1e10 * ones(nFeaturesToTry,1);
  minTrainErrorMultiNC = 1e10 * ones(nFeaturesToTry,1);
  minTrainErrorMixtureSda = 1e10 * ones(nFeaturesToTry,1);
    
  % Iterate over the number of features
  iFeatureVec = 1;
  for iFeature = 1:nFeaturesToTry
    nFeatures = nFeaturesVec(iFeature);
    
    disp(sprintf('ITERATION %g: %g FEATURES ...', iIter, nFeatures))
    
    disp('Generating data ...');
    [trainMat, trainClass, testMat, testClass] = ...
      multiCentroid_generateTrainTestData( nTrain, nTest, nFeatures, ...
      nCentroidsPerClass, flipProbs);
    
    % Compute pairwise similarity of training data and similarity of
    % test data to training data
    disp('Computing similarity matrices ...');
    [trainMatSim, testMatSim] = multiCentroid_computeSimilarityMatrices(...
      trainMat, trainClass, testMat, simFunction); 
    
    % Sort by nearest neighbor in similarity space
    testNeighborIdx = findNearNeighbors(testMatSim, maxNeighbors); 
    
    % ... Local SDA ...
    if doLocalSda == true
      disp('... local SDA ...');
      %  ... train ...
      neighborhoodSizesLocalSda = defaultNeighborhoodSizes;
      [optimumSizesLocalSda(iIter, iFeatureVec), ...
        minTrainErrorLocalSda(iIter, iFeatureVec)] = ...
        optimizeNeighborhoodLocalSda( neighborhoodSizesLocalSda, ...
        trainMatSim, trainClass);
      % ... test ...
      testClassEstimateLocalSda = multiCentroid_classifyLocalSda( ...
        optimumSizesLocalSda(iIter, iFeatureVec), testNeighborIdx, ...
        trainMatSim, testMatSim, trainClass);
      testResultLocalSda(:, iFeatureVec) = testClassEstimateLocalSda(:);
    end %if doLocalSda
    
    % ... local nearest centroid ...
    if doLocalNearestCentroid == true
      disp('... local nearest centroid ...');
      % ... train ...
      neighborhoodSizesLocalNC = defaultNeighborhoodSizes;
      [optimumSizesLocalNC(iIter, iFeatureVec), ...
        minTrainErrorLocalNC(iIter, iFeatureVec)] = ...
        optimizeNeighborhoodLocalNC(neighborhoodSizesLocalNC, ...
        trainMatSim, trainClass);
      % ... test ...
      testClassEstimateLocalNC = multiCentroid_classifyLocalNearestCentroid( ...
        optimumSizesLocalNC(iIter, iFeatureVec), testNeighborIdx, trainMatSim, ...
        testMatSim, trainClass);
      testResultLocalNC(:, iFeatureVec) = testClassEstimateLocalNC(:);
    end %if doLOcalNearestCentroid
        
    % ... Nearest Neighbor ...
    if doNearestNeighbor == true
      disp('... nearest neighbor ...');
      % .train ...
      neighborhoodSizesNN = defaultNeighborhoodSizes;
      [optimumSizesNN(iIter, iFeatureVec), ...
        minTrainErrorNN(iIter, iFeatureVec)] = optimizeNeighborhoodNN( ...
        neighborhoodSizesNN, trainMatSim, trainClass);
      % ... test ...
      neighborsToPick = testNeighborIdx(:,1:optimumSizesNN(iIter, iFeatureVec));
      neighborClass = trainClass(neighborsToPick);
      testClassEstimateNN = mode(neighborClass,2);
      testResultNN(:, iFeatureVec) = testClassEstimateNN(:);
    end %if doNearestNeighbor
      
    % ... SDA ...
    if doSda == true
      disp('... SDA ...');
      % ... train ...
      modelParamsSda = computeSdaModels(trainMatSim, trainClass, 'maxSumSim');
      % ... test ...
      testClassEstimateSda = classifySda(testMatSim, modelParamsSda);
      testResultSda(:,iFeatureVec) = testClassEstimateSda(:);
    end %if doSda
    
    % ... Mixture SDA ...
    if doMixtureSda == true
      disp('... mixture SDA ...');
      % ... train ...
      % optimize number of centroids first, then train model
      maxIter = 30;
      nCentroidsToTryVec = [2 5];
      nFolds = 5;
      [optimumNumberOfCentroidsMixtureSda{iIter, iFeatureVec}, ...
        minTrainErrorMixtureSda(iIter, iFeatureVec)] = ...
        optimizeNumberOfCentroidsMixtureSda(nCentroidsToTryVec, ...
        trainMatSim, trainClass, maxIter, nFolds);
      modelParamsMixtureSda = computeMixtureSdaModels( ...
        trainMatSim, trainClass, ...
        optimumNumberOfCentroidsMixtureSda{iIter, iFeatureVec} , ...
        'maxSumSim', maxIter);
      
      % ... test...
      testClassEstimateMixtureSda = classifyMixtureSda(...
        testMatSim, modelParamsMixtureSda);
      testResultMixtureSda(:,iFeatureVec) = testClassEstimateMixtureSda(:);
      %keyboard
    end
      
    % ... nearest centroid ...
    if doNearestCentroid == true
      disp('... nearest centroid ...')
      % ...train ...
      nearestCentroidModels = computeCentroidModels(trainMatSim, ...
        trainClass, 'maxSumSim');
      % ... test ...
      testClassEstimateNC = classifyNearestCentroid(testMatSim, ...
        nearestCentroidModels);
      testResultNC(:, iFeatureVec) = testClassEstimateNC(:);
    end %if doNearestCentroid
    
    % ... multiple nearest centroid ...
    if doMultiNearestCentroid == true
      disp('... condensed neighbor (formerly known as multiple nearest centroids) ...');
      % ... train ...
      [optimumNumberOfCentroids{iIter, iFeatureVec}, ...
        minTrainErrorMultiNC(iIter, iFeatureVec)] = ...
        optimizeNumberOfCentroids(nCentroidsToTry, trainMatSim, trainClass);
      initCentroidIdx = initMultiCentroids(...
        optimumNumberOfCentroids{iIter, iFeatureVec}, trainClass);
      multiCentroidModel = computeMultiCentroidModels(trainMatSim, ...
        trainClass, initCentroidIdx, 'maxSumSim');
      % ... test ...
      testClassEstimateMultiNC = classifyMultiNearestCentroid( ...
        testMatSim, multiCentroidModel);
      testResultMultiNC(:, iFeatureVec) = testClassEstimateMultiNC(:);
      %keyboard
    end %if doMultiNearestCentroids
    
    % ... naive Bayes ...
    if doNaiveBayes == true
      disp('... naive Bayes ...')
      % ... train ...
      naiveBayesModels = computeNaiveBayesModels(trainMat, trainClass);
      % ... test ...
      testClassEstimateNaiveBayes = classifyNaiveBayes(testMat, ...
        naiveBayesModels);
      testResultNaiveBayes(:, iFeatureVec) = testClassEstimateNaiveBayes(:);
    end %if doNaiveBayes
    
    % ... NNSim SDA ...
    if doNNSimSda == true
      disp('... NNSim SDA ...');
      % ... train ...
      %simDomainVector = 0:nFeaturesVec(iFeatureVec);
      simDomainVector = unique(unique(trainMatSim));
      modelParamsNNSimSda = computeMaxEntModelsNNSim(trainMatSim, ...
        trainClass, simDomainVector);
      % ... test ...
      [testClassEstimateNNSimSda, logPosteriorsNNSim] = ...
        classifySdaNNSim(testMatSim, trainClass, modelParamsNNSimSda);
      testResultNNSimSda(:,iFeatureVec) = testClassEstimateNNSimSda(:);
    end %if doSda
    
    iFeatureVec = iFeatureVec + 1;

  end %for iFeature
  
  % Accumulate results in matrices
  totalResultLocalSda = [totalResultLocalSda; testResultLocalSda];
  totalResultLocalNC = [totalResultLocalNC; testResultLocalNC];
  totalResultNN = [totalResultNN; testResultNN];
  totalResultSda = [totalResultSda; testResultSda];
  totalResultNC = [totalResultNC; testResultNC];
  totalResultMultiNC = [totalResultMultiNC; testResultMultiNC];
  totalResultNaiveBayes = [totalResultNaiveBayes; testResultNaiveBayes];
  totalResultMixtureSda = [totalResultMixtureSda; testResultMixtureSda];
  totalResultNNSimSda = [totalResultNNSimSda; testResultNNSimSda];
  save(fileName);
end %for iIter


%% Compute error performance
testClassMat = repmat(testClass, nIterations, length(nFeaturesVec));

if doLocalSda == true
  errVecLocalSda = sum(totalResultLocalSda ~= testClassMat);
  errVecLocalSda = errVecLocalSda / size(testClassMat,1); %#ok
end

if doSda == true
  errVecSda = sum(totalResultSda ~= testClassMat);
  errVecSda = errVecSda / size(testClassMat,1); %#ok
end

if doLocalNearestCentroid == true
  errVecLocalNC = sum(totalResultLocalNC ~= testClassMat);
  errVecLocalNC = errVecLocalNC / size(testClassMat,1);%#ok
end

if doNearestCentroid == true 
  errVecNC = sum(totalResultNC ~= testClassMat);
  errVecNC = errVecNC / size(testClassMat,1); %#ok
end
if doNearestNeighbor == true
  errVecNN = sum(totalResultNN ~= testClassMat);
  errVecNN = errVecNN / size(testClassMat,1); %#ok
end

if doNaiveBayes == true
  errVecNaiveBayes = sum(totalResultNaiveBayes ~= testClassMat);
  errVecNaiveBayes = errVecNaiveBayes / size(testClassMat,1); %#ok
end

if doMultiNearestCentroid == true
  errVecMultiNC = sum(totalResultMultiNC ~= testClassMat);
  errVecMultiNC = errVecMultiNC / size(testClassMat,1); %#ok
end

if doMixtureSda == true
  errVecMixtureSda = sum(totalResultMixtureSda ~= testClassMat);
  errVecMixtureSda = errVecMixtureSda / size(testClassMat,1); %#ok
end

if doNNSimSda == true
  errVecNNSimSda = sum(totalResultNNSimSda ~= testClassMat);
  errVecNNSimSda = errVecNNSimSda / size(testClassMat,1); %#ok
end

save(fileName);
return % END main function

%% LOCAL FUNCTIONS
% Helper local functions have been moved to separate files. Helper function
% names start with "multiCentroid_"
