function [bestSizeStruct, bestRegParamStruct, minErrStruct] = ...
  crossValidateParamsAverageLocalSda( neighborhoodSizeVec, neighborType, ...
  regParamVec, simMat, crossValIdxMat, classLabels, simDomain, testIdx)
%crossValidateParamsAverageLocalSda - cross validate parameters.
% Cross-validate the neighborhood size and the regularization parameter
% for the local SDA classifier regularized by the average local SDA pmfs.
% Uses the N-fold cross validation strategy. This function is designed to be
% used in conjucntion with readPartitionFile() and with the partition files
% from the Similarity Learning Repository.
%
% Assumes similarity data are used (not dissimilarity) and the centroid
% type is 'maxSumSim'.
%
% USAGE
% [bestSize, bestRegParam, minErr] = crossvalidateParamsAverageLocalSda( ...
%     neighborhoodSizeVec, neighborType, regParamVec, simMat, ...
%     crossValIdxMat, classLabels, simDomain, testIdx)
% neighbrohoodSizeVec: vector of neighborhood sizes to cross-validate.
% neighborType: string, can be either 'kPerClass' or 'kTotal'.
% regParamVec: vector of regularization parameters to cross-validate.
% simMat: N-by-N matrix of pairwise similarities.
% crossValIdxMat: N-by-K matrix of indices into the rows of simMat,
%    denoting the samples to be used for cross-validation. It is a cell
%    from crossValIdxCell as returned by readPartitionFile().
% classLabels: N-by-1 vector of class labels for each row of simMat.
% simDomain: vector of possible similarity values.
% testIdx: a vector of indices into the columns of simMat denoting which
%     columns should be reserved for the testing data set. It is a row from
%     testIdxMat as returned by readPartitionFile(), corresponding to
%     the matrix crossValIdxMat.
% bestSize: optimized neighborhood size.
% bestRegParam: optimized regularization parameter.
% minErr: cross-validation error rate corresponding to the optimized
%     parameters. 

% Written by Luca Cazzanti, engineer.
% Applied Physics Laboratory, University of Washington, Jan 2009.
%$Id:$
%$DateTime:$
%$Author:$

%% Preliminaries

% Specify similarity domain
if isempty(simDomain)
  simDomain = unique(simMat(:));
else
  simDomain = unique(simDomain);
end

% Vector of regularizing parameter values to test
regParamVec = unique(regParamVec);
nRegParams = length(regParamVec);

% Vector of neighborhood sizes to cross-validate 
actualNeighborhoodSizeVec = unique(neighborhoodSizeVec);

% Number of cross-validation folds
nFolds = size(crossValIdxMat,1);

% Get rid of self-similarities: we don't really need to order all other
% elements right now. Later, inside the loops, we get the neighbors
% that we need. neighborIdx is used throughout this function.
%neighborIdx = findNearNeighbors(simMat, size(simMat,2), 'similarity');
%neighborIdx = adjustSelfSimilarityRank(neighborIdx);
%neighborIdx(:,1) = [];

% END of preliminary stuff

%% Pre-compute the average local SDA parameters for each fold
% If the regularizing parameter is a scalar = 0, then we are not
% regularizing, so there is no need to compute the average local SDA models.

if nRegParams == 1
  skipAverageModels = regParamVec == 0;
else
  skipAverageModels = false;
end

% Pre-allocate
averageParamStorage = cell(nFolds, 1);

% Do not run average models
if skipAverageModels == true
  % Create a cell array of dummy sda parameter structures to be used later
  for iFold = 1:nFolds
    tmpParams = initAverageLocalSdaParams();
    tmpParams.centroidType = 'maxSumSim';
    tmpParams.simDomain = simDomain;
    tmpParams.neighborhoodSizes = actualNeighborhoodSizeVec;
    tmpParams.neighborType = neighborType;
    tmpParams.classLabels = unique(classLabels);
    nClasses = length(unique(classLabels));
    tmpParams.classPriors = 1/nClasses * ones(nClasses,1);
    tmpParams.avgMarginalConditionalLocalSdaPmfs = cell(nClasses, nClasses, max(neighborhoodSizeVec));
    tmpParams.avgSimToCentroidHistPmfs = cell(length(actualNeighborhoodSizeVec),1);
    tmpParams.avgAcrossClasses = zeros(nClasses, max(neighborhoodSizeVec));
        
    averageParamStorage{iFold} = tmpParams;
  end % for iFold

else % run average models
  
  for iFold = 1:nFolds
    disp(['Pre-computing average models for fold: ' num2str(iFold)])

    % Get the cross-validation - that is, the "test" sample for this fold
    crossValIdx = crossValIdxMat(iFold,:);
    crossValIdx(crossValIdx == 0) = []; % take out any remaining filler values

    % Compute the average local SDA distribution using the rest of the
    % samples as training samples for this cross-val fold ...

    % ... take out the true test samples ...
    thisFoldTrainIdx = 1:size(simMat,2);
    thisFoldTrainIdx(testIdx) = [];

    % ... take this fold's "test" points out of this fold's training data ...
    nCrossVals = length(crossValIdx);
    for iCrossVal = 1:nCrossVals
      idx = find(thisFoldTrainIdx == crossValIdx(iCrossVal));
      thisFoldTrainIdx(idx) = [];
    end
    thisFoldTrainSimMat = simMat(:, thisFoldTrainIdx);
    thisFoldTrainSimMat = thisFoldTrainSimMat(thisFoldTrainIdx,:);
    thisFoldTrainClassLabels = classLabels(thisFoldTrainIdx);

    thisFoldAverageLocalSdaModelParams = computeAverageLocalSdaModels( ...
      thisFoldTrainSimMat, thisFoldTrainClassLabels, 'maxSumSim', ...
      actualNeighborhoodSizeVec, simDomain, neighborType);

    % The number of tested neighborhhod sizes can decrease, but that's ok
    % because we will ultimately only use the smallest size.
    actualNeighborhoodSizeVec = intersect(actualNeighborhoodSizeVec, ...
      thisFoldAverageLocalSdaModelParams.neighborhoodSizes);

    averageParamStorage{iFold} = thisFoldAverageLocalSdaModelParams;
  end % for iFold

end %if-else skipAverageModels == true

% Number of different neighborhoods to try
nSizes = length(actualNeighborhoodSizeVec);


%% Cross-validation loops
avgMarginalErrMat = 0;
avgMarginalUniformPriorsErrMat = 0;
avgJointErrMat = 0;
avgJointUniformPriorsErrMat = 0;
avgSimToCentroidsErrMat = 0; 
avgAcrossClassesErrMat = 0;
    
iterCount = 0;
for iFold = 1:nFolds %%DEBUG
  
  disp(['Cross-validation fold: ' num2str(iFold)]);
 
  % The average local SDA model aprameters to use for this fold
  thisFoldAverageLocalSdaModelParams = averageParamStorage{iFold};
  
  % Get the cross-validation - that is, the "test" sample for this fold
  crossValIdx = crossValIdxMat(iFold,:);
  crossValIdx(crossValIdx == 0) = []; % take out any remaining filler values

  % Iterate through all the "test" samples in this cross-validation fold
  nCrossVals = length(crossValIdx);
  for iCrossVal = 1:nCrossVals

    % get the ith "test" sample's index
    thisCrossValSampleIdx = crossValIdx(iCrossVal);
    thisCrossValClass = classLabels(thisCrossValSampleIdx);

    % Remove the sample itself from candidate neighbors indeces.
    candidateNeighborIdx = 1:size(simMat,2);
    candidateNeighborIdx(thisCrossValSampleIdx) = [];
    
    % Remove other cross validation "test" points as candidate neighbors
    % by looking through each of them and removing them
    for iCount = 1:nCrossVals
      idx = find( candidateNeighborIdx == crossValIdx(iCount) );
      candidateNeighborIdx(idx) = [];
    end % for iCount

    % Remove the true test samples as candidate neighbors
    % by looking through each of them and removing them
    for iTest = 1:length(testIdx)
      idx = find( candidateNeighborIdx == testIdx(iTest) );
      candidateNeighborIdx(idx) = [];
    end % for iTest

    % Compose matrix of candidate neighbor similarities
    candidateNeighborClassLabels = classLabels(candidateNeighborIdx);
    candidateNeighborSimMat = simMat( :, candidateNeighborIdx );
    candidateNeighborSimMat = candidateNeighborSimMat( candidateNeighborIdx, : );
    testPointSimToCandidates = simMat( thisCrossValSampleIdx, candidateNeighborIdx );
   
    % Get the actual neighbors
    if strcmpi(neighborType, 'kPerClass')
      neighborIdxByClass = findNearNeighborsByClass(testPointSimToCandidates, ...
        size(candidateNeighborSimMat,2), candidateNeighborClassLabels, 'similarity');
          % There should be only one element in the cell array, a matrix.
          neighborIdxMat = neighborIdxByClass{1};
    elseif strcmpi(neighborType, 'kTotal')
      neighborIdxMat = findNearNeighbors(testPointSimToCandidates, ...
        size(candidateNeighborSimMat,2));
    else
      error(['Unsupported neighborhood type: ' neighborType]);
    end
      
    % Iterate over the neighborhood sizes
    avgMarginalClassEstMat = -1e6*ones(nSizes, nRegParams);
    avgMarginalUniformPriorsClassEstMat = -1e6*ones(nSizes, nRegParams);
    avgJointClassEstMat = -1e6*ones(nSizes, nRegParams);
    avgJointUniformPriorsClassEstMat = -1e6*ones(nSizes, nRegParams);
    avgSimToCentroidsClassEstMat = -1e6*ones(nSizes, nRegParams);
    avgAcrossClassesClassEstMat = -1e6*ones(nSizes, nRegParams);

    for iSize = 1:nSizes
      
      nNeighbors = actualNeighborhoodSizeVec(iSize);
      actualNeighborIdx = neighborIdxMat(:, 1:nNeighbors);
      actualNeighborIdx = actualNeighborIdx(:).';

      actualTestPointSim = testPointSimToCandidates(actualNeighborIdx);
      actualNeighborSimMat = candidateNeighborSimMat(:, actualNeighborIdx);
      actualNeighborSimMat = actualNeighborSimMat(actualNeighborIdx, :);
      actualNeighborClassLabels = candidateNeighborClassLabels(actualNeighborIdx);

      % Iterate over the possible values of the reg parameter
      classEst =  -1e6 * ones(1, nRegParams);
      classEst2 = -1e6 * ones(1, nRegParams);
      for iParam = 1:nRegParams

        regularizingParams = createRegularizingParamStruct( ...
          thisFoldAverageLocalSdaModelParams, nNeighbors, regParamVec(iParam) ) ;
      
        classEstStruct = classifyRegularizedLocalSda(...
          actualTestPointSim, actualNeighborSimMat, ...
          actualNeighborClassLabels, regularizingParams);
        
        % Put results from various regularization strategies in separate
        % variables
        avgMarginalClassEst(iParam) = classEstStruct.avgMarginal;
        avgMarginalUniformPriorsClassEst(iParam) = classEstStruct.avgMarginalUniformPriors;
        avgJointClassEst(iParam) = classEstStruct.avgJoint;
        avgJointUniformPriorsClassEst(iParam) = classEstStruct.avgJointUniformPriors;
        avgSimToCentroidsClassEst(iParam)= classEstStruct.avgSimToCentroids;
        avgAcrossClassesClassEst(iParam) = classEstStruct.avgAcrossClasses;
        
      end % for iParam

    avgMarginalClassEstMat(iSize,:) = avgMarginalClassEst;
    avgMarginalUniformPriorsClassEstMat(iSize,:) = avgMarginalUniformPriorsClassEst;
    avgJointClassEstMat(iSize,:) = avgJointClassEst;
    avgJointUniformPriorsClassEstMat(iSize,:) = avgJointUniformPriorsClassEst;
    avgSimToCentroidsClassEstMat(iSize,:) = avgSimToCentroidsClassEst;
    avgAcrossClassesClassEstMat(iSize,:) = avgAcrossClassesClassEst;
      
      %DEBUGGING
      %[tmpClassEst(iSize) tmpLogPosteriors] = classifyLocalSda(...
      %  actualTestPointSim, actualNeighborSimMat, ...
      %  actualNeighborClassLabels, 'maxSumSim', simDomain);
      % END DEBUGGING
      
      
    end %for iSize

    
    % Accumulate the count of incorrect classifications for all samples in
    % all cross validation folds
    thisCrossValErrMat = double(avgMarginalClassEstMat ~= thisCrossValClass);
    avgMarginalErrMat = avgMarginalErrMat + thisCrossValErrMat;

    thisCrossValErrMat = double(avgMarginalUniformPriorsClassEstMat ~= thisCrossValClass);
    avgMarginalUniformPriorsErrMat = avgMarginalUniformPriorsErrMat + thisCrossValErrMat;
    
    thisCrossValErrMat = double(avgJointClassEstMat ~= thisCrossValClass);
    avgJointErrMat = avgJointErrMat + thisCrossValErrMat;
    
    thisCrossValErrMat = double(avgJointUniformPriorsClassEstMat ~= thisCrossValClass);
    avgJointUniformPriorsErrMat = avgJointUniformPriorsErrMat + thisCrossValErrMat;
    
    thisCrossValErrMat = double(avgSimToCentroidsClassEstMat ~= thisCrossValClass);
    avgSimToCentroidsErrMat = avgSimToCentroidsErrMat + thisCrossValErrMat;
    
    thisCrossValErrMat = double(avgAcrossClassesClassEstMat ~= thisCrossValClass);
    avgAcrossClassesErrMat = avgAcrossClassesErrMat + thisCrossValErrMat;
    
    iterCount = iterCount + 1;
    
  end % for iCrossVal

end % for iFold

avgMarginalErrMat = avgMarginalErrMat / iterCount;
avgMarginalUniformPriorsErrMat = avgMarginalUniformPriorsErrMat / iterCount;
avgJointErrMat = avgJointErrMat / iterCount;
avgJointUniformPriorsErrMat = avgJointUniformPriorsErrMat / iterCount;
avgSimToCentroidsErrMat = avgSimToCentroidsErrMat / iterCount;
avgAcrossClassesErrMat = avgAcrossClassesErrMat / iterCount;


% Get the parameter values corresponding to the minimum error
[bestSize, bestRegParam, minErr] = local_findBestParams(avgMarginalErrMat, ...
  actualNeighborhoodSizeVec, regParamVec);
bestSizeStruct.avgMarginal = bestSize;
bestRegParamStruct.avgMarginal = bestRegParam;
minErrStruct.avgMarginal = minErr;

[bestSize, bestRegParam, minErr] = local_findBestParams(avgMarginalUniformPriorsErrMat, ...
  actualNeighborhoodSizeVec, regParamVec);
bestSizeStruct.avgMarginalUniformPriors = bestSize;
bestRegParamStruct.avgMarginalUniformPriors = bestRegParam;
minErrStruct.avgMarginalUniformPriors = minErr;

[bestSize, bestRegParam, minErr] = local_findBestParams(avgJointErrMat, ...
  actualNeighborhoodSizeVec, regParamVec);
bestSizeStruct.avgJoint = bestSize;
bestRegParamStruct.avgJoint = bestRegParam;
minErrStruct.avgJoint = minErr;

[bestSize, bestRegParam, minErr] = local_findBestParams(avgJointUniformPriorsErrMat, ...
  actualNeighborhoodSizeVec, regParamVec);
bestSizeStruct.avgJointUniformPriors = bestSize;
bestRegParamStruct.avgJointUniformPriors = bestRegParam;
minErrStruct.avgJointUniformPriors = minErr;

[bestSize, bestRegParam, minErr] = local_findBestParams(avgSimToCentroidsErrMat, ...
  actualNeighborhoodSizeVec, regParamVec);
bestSizeStruct.avgSimToCentroids = bestSize;
bestRegParamStruct.avgSimToCentroids = bestRegParam;
minErrStruct.avgSimToCentroids = minErr;

[bestSize, bestRegParam, minErr] = local_findBestParams(avgAcrossClassesErrMat, ...
  actualNeighborhoodSizeVec, regParamVec);
bestSizeStruct.avgAcrossClasses = bestSize;
bestRegParamStruct.avgAcrossClasses = bestRegParam;
minErrStruct.avgAcrossClasses = minErr;

return % end main function

%% LOCAL FUNCTIONS
%local_findBestParams()
function [bestSize, bestRegParam, minErr] = ...
  local_findBestParams( errMat, neighborhoodSizeVec, regParamVec)

[minErrTemp, minErrSizeIdx] = min(errMat);
[minErr, minErrRegParamIdx] = min(minErrTemp);
bestRegParam = regParamVec(minErrRegParamIdx);
try
bestSize = neighborhoodSizeVec(minErrSizeIdx(minErrRegParamIdx));
catch
  lasterr
  keyboard
end
return % END local_findBestParas()
