function [classEst, loglike] = classifyNaiveBayes(dataMat, modelParams)
%classifyNaiveBayes - classify using the naive Bayes classifier.
% Classifies data using a naive Bayes classifier. The naive Bayes
% probabilities are assumed previously computed by
% computeNaiveBayesModels(). This function applies the Laplace correction
% to the probability estimates of both the class priors and the
% class-conditional feature pmfs.
%
% USAGE
% [classEst, loglike] = classifyNaiveBayes(dataMat, modelParams)
% dataMat: N-by-M matrix of data to be classified; N samples, each with M
%     features.
% modelParams: structure containing the probability models as computed by
% computeNaiveBayesModels().
% classEst: N-by-1 vector of estimated class labels; the i-th number
%     corresponds to the most likely model for the i-th row of dataMat.
% loglike: N-by-C matrix of class-conditional log-likelihoods of the data
%     given C class models.
%
% See also initNaiveBayesModels, comuteNaiveBayesModels.

% Written by Luca Cazzanti, engineer
% Applied Physics Laboratory, University of Washington, Dec 2006.
%$Id:$
%$DateTime:$
%$Author:$

% These are used later on
[nData, notUsed] = size(dataMat); %#ok 

%% Trivial model --> Return the only class as the class estimate, done.
% This code is written more generally than the one in
% computeNaiveBayesModels(); it allows for the multiclass case
% classPriors = [0 1 0 0], whereas for the trivial case
% computeNaiveBayesModels() always produces classPriors = [1]. It works
% either way.
if modelParams.isTrivial == true
  theOnlyClassLabelIdx = find(modelParams.classPriors == 1);
  if isempty(theOnlyClassLabelIdx)
    error('A trivial model was provided, but no class prior = 1');
  end
  theOnlyClassLabel = modelParams.classlabels(theOnlyClassLabelIdx);
  classEst = theOnlyClassLabel * ones(nData,1);
  loglike = -Inf * ones(nData, length(modelParams.classPriors));
  loglike(:,theOnlyClassLabelIdx) = 0;
  return
end

%% Non-trivial case
% Build matrix of class-conditional likelihoods: i-th column is the
% likelihood of the data point conditioned on the i-th class.
[nClasses, nFeatures] = size(modelParams.probModels);
loglike = zeros(nData, nClasses);
for iClass = 1:nClasses
  probMat = zeros(nData, nFeatures);
  for iFeature = 1:nFeatures
    possibleFeatureVals = modelParams.featureVals{iClass, iFeature};
    probVals = modelParams.probModels{iClass, iFeature};
    for iFeatureValue = 1:length(possibleFeatureVals)
      thisValue = possibleFeatureVals(iFeatureValue);
      idx = find( dataMat(:, iFeature) == thisValue );
      probMat(idx, iFeature) = probVals(iFeatureValue); %#ok
    end %for iFeatureValue
  end %for iFeature
  
  % Laplace correction
  probMat = local_laplaceCorrection( probMat, ...
    modelParams.featureVals(iClass,:), modelParams.classCounts(iClass) );
  
  loglike(:, iClass) = sum(log(probMat),2);
  
end %for iClass

% Take into account class priors and classify
loglike = loglike + repmat(log(modelParams.classPriors(:).'), nData, 1);

% Classify
[notUsed, classEstIdx] = max(loglike, [], 2); %#ok
classEst = modelParams.classLabels(classEstIdx);

return %END main function

%% LOCAL FUNCTIONS
function probMat = local_laplaceCorrection(probMat, classFeatureVals, ...
  classSize)
% Apply Laplace correction to class-conditional feature probability
% estimates.

% Debugging hint: it must be nFeatures = size(probMat,2);
nFeatures = length(classFeatureVals);

for iFeature = 1:nFeatures
  nFeatureVals = length(classFeatureVals{iFeature});
  probMat(:,iFeature) = ( probMat(:,iFeature) * classSize + 1 ) / ...
    ( classSize + nFeatureVals);
end %for iFeature

return %END local_laplaceCorrection()
      