function modelParams = computeBdaModels(featureMat, classLabels)
%computeBdaModels - compute model parameters for BDA.
% Computes the parameters of the
%
%USAGE 
% modelParams = computeBdaModels(featureMat, classLabels)
% featureMat: N-by-F matrix, each row containing the F features that
% describe a sample.
% classLabels: M-by-1 vector of class labels for the M samples.
% modelParams: structure containing the BDA model parameters. The fields are:
%     Bh: the Bh term in the BDA classifier. 
%     classCounts: number of samples per class.
%     classPriors: class prior probabilities, estimated with Laplace
%     correction.
%     q: parameter for the inverse Wishart prior.
%     classCovNoNorm: G-by-1 cell array. The g-th cell contains the F-by-F
%     normalized class-conditional covariance mattrix for class g samples.
%     classMeans: G-by-F matrix of class mean vectors.
%     gammaTerm: the log-Gamma term in the BDA classifier.
%     classLabels: G-by-1 vector of possible class labels. 
%
% See also classifyBda, classifyLocalBda.

% Written by Luca Cazzanti, enginner.
% Applied Physics Lab - University of Washington, Feb 2009.
% Originally written by Santosh Srivastava and Sergey Feldman,
% University of Washington, Dept. Electrical Engineering.


% Some parameters used throughout
d = size(featureMat,2); % dimension of feature space
uniqueClassLabels = unique(classLabels);
nClasses = length(uniqueClassLabels);

% Should be the same number of neighbors for each class
numPerClassVec = hist(classLabels, uniqueClassLabels);
%numPerClass = max(numPerClassVec); % used later

% Scale parameter for the inverted wishart prior
q = d + 3; 

% Compute the local covariances, means, and other BDA parameters.

pooledCovNoNorm = zeros(d,d);
classMeans = zeros(nClasses, d);
classCov = cell(nClasses,1);
classCovNoNorm = cell(nClasses,1);
for iClass = 1:nClasses
  idx = find(classLabels == uniqueClassLabels(iClass));
  thisClassFeatureMat = featureMat(idx,:);
  classMeans(iClass,:) = mean(thisClassFeatureMat,1);
  
  % Is there only one neighbor for each class? Then special care for
  % covariances
  if numPerClassVec(iClass) > 1 
    classCov{iClass} = cov(thisClassFeatureMat);
    classCovNoNorm{iClass} = classCov{iClass} * (numPerClassVec(iClass) - 1); 
  else
    classCov{iClass} = zeros(d,d);
    classCovNoNorm{iClass} = zeros(d,d);
  end % if neighborsPerClass > 1
    pooledCovNoNorm = pooledCovNoNorm + classCovNoNorm{iClass};
end
pooledCov = pooledCovNoNorm / ( sum(numPerClassVec) - 1 ); 

% BDA parameters
diagCov = diag(pooledCov);
Bh = .95*q*diag(diagCov) + .05*eye(d); % Bh term
% We can precalculate the gamma term because we take the same
% number of samples from each class. See equation (7) in Completely
% Lazy Classifiers by Garcia, Feldman, Gupta, and Srivastava
gammaTerm = gammaln(0.5*(numPerClassVec+q+1)) - ...
  gammaln(0.5*(numPerClassVec+q-d+1));

modelParams.Bh = Bh;
modelParams.classCounts = numPerClassVec;
modelParams.classPriors = (modelParams.classCounts + 1) / (sum(modelParams.classCounts) + nClasses);
modelParams.q = q;
modelParams.classCovNoNorm = classCovNoNorm;
modelParams.classMeans = classMeans;
modelParams.gammaTerm = gammaTerm;
modelParams.classLabels = uniqueClassLabels;
