function [sortIdx, sortMat] = findNearNeighbors(inMat, nNeighbors, varargin)
%findNearNeighbors - find the nearest neighbors.
% Finds the indices of a sample's nearest neighbors, given a matrix of
% pairwise (dis)similarities.
%
% USAGE 
% [neighborIdx, neighborSim] = findNearNeighbors(inMat, nNeighbors, simMode)
% inMat - N-by-M matrix of (dis)similarities; each row contains a sample's
%     (dis)similarities to M samples.
% nNeighbors - number of neighbors to return
% simMode - optional, either 'similarity' (default) or 'dissimilarity',
%     depending on whether inMat contains similarities or dissimilarities.
% neighborIdx - N-by-nNeighbors matrix; each row contains the column
%     indices of inMat corresponding to that sample's nearest neighbors.
% neighborSim - N-by-nNeighbors matrix of similarities between each sample
%     and its nearest neighbors: neighborSim = inMat(:,neighborIdx).
%
% See also adjustSelfSimilarityRank

% Written by Luca Cazzanti, engineer
% Applied Physics Laboratory - University of Washington, Dec. 2006.
%$Id:$
%$DateTime:$
%$Author:$

% Similarities or dissimilarities?...
switch nargin
  case 2 % default is similarities --> descend
    sortMode = 'descend';
  
  case 3
    switch varargin{1}
      case 'similarity'
        sortMode = 'descend';
      case 'dissimilarity'
        sortMode = 'ascend';
      otherwise
        error('Invalid similarity type');
    end % switch inMode
  
  otherwise
    error('Wrong number of inputs');
end % switch nargin

% Sort similarities
[sortMat, sortIdx] = sort(inMat, 2, sortMode);

% Pick only up to nNeighbors columns, ... 
maxNumNeighbors = min( [nNeighbors size(inMat,2)] );
sortIdx = sortIdx(:,1:maxNumNeighbors);
sortMat = sortMat(:,1:maxNumNeighbors);

% CAREFUL: might need to run adjustSelfSimilarityRank() after running this
% function. We want to ensure sortIdx(rowIdx,1) = rowIdx, that is  each
% element is its own nearest neighbor even if it is equally similar to
% other elements of the row. Matlab's sort() does not do this; it maintains
% the original sort order if there are (dis)similarity ties so we might
% need to adjust the neighbors ranks.
% This is accomplished by the separate function adjustSelfSimilarityRank().

return
